{% if show %}
<nav class="toplinks">
{% block date-hierarchy-toplinks %}
{% block date-hierarchy-back %}
{% if back %}<a href="{{ back.link }}" class="date-back">&lsaquo; {{ back.title }}</a>{% endif %}
{% endblock %}
{% block date-hierarchy-choices %}
{% for choice in choices %}
{% if choice.link %}<a href="{{ choice.link }}">{{ choice.title }}</a>{% else %}{{ choice.title }}{% endif %}
{% endfor %}
{% endblock %}
{% endblock %}
</nav>
{% endif %}
